/* 
 * CSE142, Spring 1999
 * Homework 4
 * Traffic animation

  STARTER PROGRAM

 * Section BH
 *
 */


#include "gp142.h"
#include <stdio.h>
#include <math.h>

#include <time.h>
	/* 
	* <time.h> is included so the current time can be accessed
	* to initialize the random number generator.  If you don't
	* use random numbers (and you don't have to) you can either
	* ignore this #include or delete it.
	*/

/* Logical constants */
#define TRUE    		1
#define FALSE   		0

/* 
 * For many of the GP142 functions, a line width of
 * zero means to fill the shape being drawn.  The
 * constant FILL is used in function calls when a solid
 * shape is desired.
 */
#define FILL 0


/*
 * Symbolic constants.  Change as needed.
 */

/* Menu buttons: */

/* Top-left coordinates of button palette */
#define MENU_TOP		(GP142_YMAX - 5)
#define MENU_LEFT		(-GP142_XMAX + 5)

/* Height and width of buttons in pixels */
#define BUTTON_HEIGHT  	25
#define BUTTON_WIDTH 	85

/* Number of buttons */
#define NUM_BUTTONS 	7


/* Possible locations of mouse clicks: */

/*   Mouse click in some button */
#define BUTTON_CLICK 1
#define ROAD_CLICK 2

/*   Mouse click elsewhere */
#define OTHER_CLICK 0


/* Coordinates of objects in the drawing */

/* y-coordinate of the top of the ground in the landscape */
#define HORIZON (- 2 * (GP142_YMAX / 10))

/* Vertical center and width of highway */
#define HIGHWAY_CENTER ((-GP142_YMAX + HORIZON) / 3)
#define HIGHWAY_WIDTH  10


/* Maximum number of vehicles that can appear in the animation */
#define MAX	20

/* Height and width of a single car */
#define CAR_HEIGHT 24
#define CAR_WIDTH  40

/* Height and width of a single monster truck*/
#define MONSTER_HEIGHT 50
#define MONSTER_WIDTH 55

/* Switches sunset to sunrise or vice versa */
#define SUNRISE 1
#define SUNSET -1
/* stages the sun goes through */
#define STAGES 6


/*****************************************************************************

     FUNCTION PROTOTYPES

*****************************************************************************/

/* Initialize random number generator.  [Ignore or		*/
/* remove if your program doesn't use random numbers.]	*/
void initialize_random_numbers();

/* Draw the background, including the sun at coordinates sun_x and sun_y */
void draw_landscape(int sun_x, int sun_y);

/* Draw the buttons */
void draw_buttons(int sun_y);

/* Return kind of mouse click (in a button, etc.) */
int  classify_mouse_click(int mouse_x, int mouse_y);

/* Process button click at location x, y.  Update car speed,  */
/* direction, color, and whether traffic is moving as needed. */
void handle_button(int x, int y, int traffic_moving[MAX], int car_speed[MAX],
				 int car_direction[MAX], int car_color[MAX], 
				 int *sun_height, int num_cars, int car_num, int car_y[MAX]);

/* Update sun coordinates sun_x, sun_y for one tick */
void move_sun(int *sun_x, int *sun_y, int sun_height);

/* Update all vehicle x, y coordinates, given current x, y */
/* location, direction (+1 or -1), and speed in pixels per tick. */
void move_traffic(int x[MAX], int y[MAX], int direction[MAX], 
				  int speed[MAX], int num_cars, int i);

/* Draw all vehicles given their current x, y coordinates and color */
void draw_traffic(int x[MAX], int y[MAX], int color[MAX], int sun_y, 
				  int car_direction[MAX], int num_cars);

/* Draw a single car of the given color with horizontal */
/* center x and bottom y*/
void draw_car(int x[MAX], int y[MAX], int color[MAX], int sun_y, 
			  int direction[MAX], int i);

/* Display error message if internal error is detected. */
void something_is_wrong (void);

int sun_position(int sun_y);

void initial_info(int y[MAX], int speed[MAX], int direction[MAX],
				  int moving[MAX], int color[MAX], int i);

void traffic(int car_direction[MAX], int x[MAX], int y[MAX], 
			 int num_cars);

/*****************************************************************************

     MAIN

*****************************************************************************/

/*
 * Main program contents:
 *   - Major program variables that persist throughout execution
 *   - Event loop that processes GP142 events until the user quits,
 *     and updates variables and redraws the window as needed.
 */
int main(void)
{
	/* Event handling and buttons: */
	int  quit;				/* = "user has selected quit"				*/
	int  event;				/* the most recent GP142 event				*/
	char key_pressed;		/* last keyboard character from key event	*/
	int  mouse_x, mouse_y;	/* x and y coordinates of latest mouse event*/
	int  click_kind;		/* kind of mouse click (button, etc.)		*/

	/* Information about the sun: */
	int sun_x, sun_y;			/* screen x and y coordinates */
	int sun_height = SUNRISE;   /* the sun is currently up */

	/* Information about vehicles in the animation:							*/		
	int num_cars = 1;		/* how many cars are on the road	*/
	int car_direction[MAX];	/* current direction: +1=right, -1=left			*/								
	int car_speed[MAX];		/* speed in pixels per tick when traffic moving	*/	
	int traffic_moving[MAX];	/* = "traffic is currently moving"	*/
	int car_x[MAX], 
		car_y[MAX];			/* current position of the car		*/
	int car_color[MAX];		/* current color of the car			*/
	int i;					/* specifies which car to update	*/
	int car_num = -1;	    /* which car is selected.  -1 is all cars */

	char n;					/* used for keyboard loop */
	int m;

	initialize_random_numbers();
	
	/* Initialize graphics package, clear the window,	*/
	/* turn logging off, and start animation.			*/
	GP142_open();
	GP142_logging(LOG_OFF);
	GP142_clear();
	GP142_animate(ANI_RUN);

	/* set initial location of the sun */
	sun_x = GP142_XMAX - 50;
	sun_y = GP142_YMAX - 35;

	/* set initial conditions for the first car */
	i = 0;
	car_x[0] = 0;
	initial_info(car_y, car_speed, car_direction, traffic_moving,
		car_color, i);
	
	/* Draw initial landscape, buttons, and traffic */
	draw_landscape(sun_x, sun_y);
	draw_buttons(sun_y);
	draw_traffic(car_x, car_y, car_color,
		sun_y, car_direction, num_cars);
	

	/*
	 * Main event loop:
	 * ---- ----- -----
	 * Wait for the next user action, decode it, and call an
	 * appropriate function to handle it.  Repeat until the
	 * user selects quit.
	 */
	quit = FALSE;
	while (!quit) {
		
		/* get next event */
		event = GP142_await_event(&mouse_x, &mouse_y, &key_pressed);

		/* perform appropriate action depending on kind of event */
		switch (event) {
		case GP142_QUIT:
			/* Quit selected */
			quit = TRUE;
			break;
		case GP142_KBD:
			/* Keyboard click.  Assign the number typed in  */
			/* to a car and process accordingly.			*/
			m = -1;
			for (n = '0'; n <= key_pressed; n = n + 1) {
				car_num = m; 
				m++;
			}
			break;
		case GP142_MOUSE:
			/* Mouse click.  Determine if in a menu button	*/
			/* or elsewhere and process accordingly.		*/
			click_kind = classify_mouse_click(mouse_x, mouse_y);
			if (click_kind == BUTTON_CLICK) {
				handle_button(mouse_x, mouse_y, traffic_moving, car_speed,
							  car_direction, car_color, &sun_height,
							  num_cars, car_num, car_y);
			} else if (click_kind == ROAD_CLICK) {
				/* Adds a car if user clicks on the road */
				i = num_cars;
				num_cars = num_cars + 1;
				car_x[i] = mouse_x;
				initial_info(car_y, car_speed, car_direction, 
							 traffic_moving, car_color, i);
			}
			break;
			
		case GP142_PERIODIC:
			/* Timer interval event.  Update everything	*/
			/* that needs to change as time advances.	*/
			move_sun(&sun_x, &sun_y, sun_height);
			for (i = 0; i < num_cars; i++) {
				if (traffic_moving[i])
					move_traffic(car_x, car_y, car_direction,
					car_speed, num_cars, i);
			}
			break;
			
		default:
			/* unknown event */
			break;
			
		}  /*end switch */
		
		/* Redraw everything */
		draw_landscape(sun_x, sun_y);
		draw_buttons(sun_y);
		/*draw_traffic(car_x, car_y, car_color, sun_y, car_direction);*/
		for (i = 0; i < num_cars; i++) {
			draw_traffic(car_x, car_y, car_color, sun_y, 
				car_direction, num_cars);
		}

	}  /* end event loop */
	
	
	/* Termination: shut down graphics window and exit */
	GP142_close();
	return 0;
}  /*end main */


/*****************************************************************************

     BACKGROUND LANDSCAPE

*****************************************************************************/


/* Draw the background, including the sun at coordinates sun_x and sun_y */
void draw_landscape(int sun_x, int sun_y)
{
	int text_offset = -GP142_XMAX + 15;	/* horizontal offset of text */
	int text_size   = 12;				/* size of text font */
	int i;								/* Height of the sun */
	int sky_color[STAGES]= {ICE_BLUE, CYAN, PEACH, MAGENTA, PURPLE, BLACK};
	int sun_color[STAGES]= {YELLOW, YELLOW, ORANGE, RED, RED, RED};

	/* find out which of the six heights the sun is at */
	i = sun_position(sun_y);

	/* Draw the sky */
	GP142_rectangleXY(sky_color[i], -GP142_XMAX, -GP142_YMAX,
		GP142_XMAX, GP142_YMAX, FILL);
	
	/* Draw the sun */
	GP142_circleXY(sun_color[i], sun_x, sun_y , 20);
	
	/* Draw the ground, highway, and text legend */
	GP142_rectangleXY(BROWN, -GP142_XMAX, -GP142_YMAX,
		GP142_XMAX, HORIZON, FILL);
	GP142_rectangleXY(WHITE, 
					-GP142_XMAX, HIGHWAY_CENTER - HIGHWAY_WIDTH,
					 GP142_XMAX, HIGHWAY_CENTER + HIGHWAY_WIDTH,
					 FILL);
	GP142_lineXY(BLACK, -GP142_XMAX, HIGHWAY_CENTER, GP142_XMAX,
					HIGHWAY_CENTER, 1);
	
	/*Draw some mountains */
	GP142_triangleXY(MED_GRAY,
		-GP142_XMAX, HORIZON,
		GP142_XMAX, HORIZON,
		0, (4 * GP142_YMAX)/5 ,
		FILL);
	
	GP142_triangleXY(OLIVE,
		-( 5 * GP142_XMAX/4), HORIZON,
		(GP142_XMAX)/3  , HORIZON,
		-(3* GP142_XMAX)/4, (2 * GP142_YMAX)/3 ,
		FILL);
	
	GP142_triangleXY(FOREST_GREEN,
		-GP142_XMAX / 3 , HORIZON,
		GP142_XMAX, HORIZON,
		GP142_XMAX, (2 * GP142_YMAX)/3 ,
		FILL);

	/* Print instructions on screen */
	GP142_printfXY(BLACK, text_offset, HIGHWAY_CENTER * 2, 16,
		"Click on the road to add a new car.");
	GP142_printfXY(BLACK, text_offset, HIGHWAY_CENTER * 2 - 20, 16,
		"Type in the number of a car to control it. Type 0 to control all.");
	
} /*end draw_landscape */



/*****************************************************************************

     BUTTONS AND MOUSE CLICKS

*****************************************************************************/


/* Draws the buttons */
void draw_buttons(int sun_y)
{
	int i;							
	int text_offset = MENU_LEFT+10;	/* offset of button text in button */
	int font_size = 12;				/* size of button text */
	int color = BLACK;
	
	if (sun_position(sun_y) == 5) color = WHITE;

	/* draw rectangles for each button */
	for(i = 0; i < NUM_BUTTONS; i++) {
		GP142_rectangleXY(color,
			MENU_LEFT,
			MENU_TOP - i * BUTTON_HEIGHT,
			MENU_LEFT + BUTTON_WIDTH,
			MENU_TOP - (i + 1)*BUTTON_HEIGHT, 1);
	}

	/* display button text */
	i = MENU_TOP - (BUTTON_HEIGHT/2) - (font_size/2);
	GP142_printfXY(FOREST_GREEN, text_offset, i, font_size, "Go");	
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(RED, text_offset, i, font_size, "Stop");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(color, text_offset, i, font_size, "Faster");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(color, text_offset, i, font_size, "Slower");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(color, text_offset, i, font_size, "Reverse");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(color, text_offset, i, font_size, "New Color");

	i -= BUTTON_HEIGHT;
	GP142_printfXY(color, text_offset, i, font_size, "Sunset/Sunrise");
	
}

/* Classify mouse click at (mouse_x,mouse_y) and return kind */
int  classify_mouse_click(int mouse_x, int mouse_y)
{
	if ((mouse_x > MENU_LEFT) &&
		(mouse_x < MENU_LEFT + BUTTON_WIDTH) &&
		(mouse_y < MENU_TOP) &&
		(mouse_y > (MENU_TOP - NUM_BUTTONS*BUTTON_HEIGHT)))
			return BUTTON_CLICK;
	else if ((mouse_y > HIGHWAY_CENTER - (HIGHWAY_WIDTH / 2)) &&
			 (mouse_y < HIGHWAY_CENTER + (HIGHWAY_WIDTH / 2)))
			return ROAD_CLICK;
	else return OTHER_CLICK;
}



/* Process button click at location x, y.  Update car speed,	*/
/* direction, color, and whether traffic is moving as needed.	*/
void handle_button(int x, int y, int traffic_moving[MAX], int car_speed[MAX],
				   int car_direction[MAX], int car_color[MAX], 
				   int *sun_height, int num_cars, int car_num, int car_y[MAX])
{
	int buttonNum;		/* number of selected button */
	int i;				/* which car to update */

	/* Complain if x coordinate couldn't be inside a button */
	if (x < MENU_LEFT || x > MENU_LEFT + BUTTON_WIDTH) {
		something_is_wrong();
		return;
	}
	
	/* Convert y coordinate to button number from top. */
	buttonNum = (MENU_TOP - y)/BUTTON_HEIGHT;	 
	
	for (i=0; i < num_cars; i++) {
		if (i == car_num || car_num == -1) {
			switch(buttonNum) {
			case 0:
				/* go */
				traffic_moving[i] = TRUE;
				break;
			case 1:
				/* stop */
				traffic_moving[i] = FALSE;
				break;
		
			case 2:
				/* speed up */
				car_speed[i] = car_speed[i] + 1;
				break;
			case 3:
				/* slow down */
				car_speed[i] = car_speed[i] - 1;
				if (car_speed[i] < 0)
					car_speed[i] = 0;
				break;
		
			case 4:
				/* reverse */
				car_direction[i] = - car_direction[i];
				/* puts car on right side of the road */
				car_y[i] = car_y[i] + (car_direction[i] * -HIGHWAY_WIDTH);
				break;
			case 5:
				/* change car color */
				car_color[i] = (car_color[i]+1) % MAX_COLORS;
				break;
			case 6:
				/* causes sun to set or rise */
				*sun_height = -*sun_height;
				i = num_cars;
				break;
			case 7:
				/* nothing happens, loops back to top */
				break;
			default:
				/*Bad news if this is ever reached! */
				something_is_wrong();
				break;
			if (i == car_num) i = num_cars;
			}
		}
	}
}



/*****************************************************************************

     TRAFFIC AND OTHER ANIMATED OBJECTS

*****************************************************************************/


/* Update sun coordinates sun_x, sun_y by one tick */
void move_sun(int *sun_x, int *sun_y, int sun_height) {
	/* Move the sun to the left.  If it reaches the edge, move it */
	/* to the other side and down a bit. */
	(*sun_x) --;
	if (*sun_x < -GP142_XMAX -30) {
		*sun_x = - *sun_x;
		(*sun_y) --;
	}
	/* sun goes down 1 unless it is at it's lowest point */
	if (sun_height == SUNSET) {
		if (*sun_y > HORIZON + 65) *sun_y = *sun_y - 1;
	}
	/* sun goes up 1 unless it is at it's highest point */
	else if (sun_height == SUNRISE) {
		if (*sun_y < GP142_YMAX - 35) *sun_y = *sun_y + 1;
	}
}

/* Update all vehicle x, y coordinates, given current x, y */
/* location, direction (+1 or -1), and speed in pixels per tick, */
void move_traffic(int x[MAX], int y[MAX], int direction[MAX], 
				  int speed[MAX], int num_cars, int i) {

	int trial_x[MAX];		/* new value for x coordinate */

	/* move vehicle.  If it moves off the edge, returns on other side */
	trial_x[i] = x[i] + (direction[i] * speed[i]);
	if (abs(trial_x[i]) > GP142_XMAX + CAR_WIDTH)
		x[i] = -direction[i] * GP142_XMAX;
	else
		x[i] = trial_x[i];
	/* make sure cars don't rear end each other */
	traffic(direction, x, y, num_cars);
}


/* Draw all vehicles given their current x, y coordinates and color */
void draw_traffic(int x[MAX], int y[MAX], int color[MAX], int sun_y, 
				  int direction[MAX], int num_cars) {
	int i;

	/* calls draw_car for each car in the array */
	for (i = 0; i < num_cars; i++) {
		draw_car(x ,y ,color , sun_y , direction, i);
	}
}



/* Draw a single car of the given color with horizontal */
/* center x and bottom y*/
void draw_car(int x[MAX], int y[MAX], int color[MAX], int sun_y, 
			  int direction[MAX], int i) {
	GP142_rectangleXY(color[i], x[i] - CAR_WIDTH/2, y[i] + CAR_HEIGHT/4,
						x[i] + CAR_WIDTH/2, y[i] + CAR_HEIGHT*3/4, FILL);
	GP142_rectangleXY(color[i], x[i] - CAR_WIDTH/6, y[i] + CAR_HEIGHT*3/4,
						x[i] + CAR_WIDTH/6, y[i] + CAR_HEIGHT, FILL);
	GP142_circleXY(BLACK, x[i] - CAR_WIDTH/4, y[i] + CAR_HEIGHT/4, CAR_HEIGHT/4);
	GP142_circleXY(BLACK, x[i] + CAR_WIDTH/4, y[i] + CAR_HEIGHT/4, CAR_HEIGHT/4);
	/* Print the number of the car on it's side */
	GP142_printfXY(BLACK, x[i] - 2, y[i] + 5, 12, "%d", i+1);
	/* Create headlights if it is dark out */
	if (sun_position(sun_y) >= 3) {
		GP142_triangleXY(YELLOW, 
			x[i] + direction[i] * (CAR_WIDTH/2), y[i] + CAR_HEIGHT/2,
			x[i] + direction[i] * (CAR_WIDTH * 2), y[i] + CAR_HEIGHT,
			x[i] + direction[i] * (CAR_WIDTH * 2), y[i],
			FILL);
	}
}

/* determines which of the 6 stages the sun is at */
int sun_position(int sun_y)
{
	double sun_pos;
	int isun_pos;
	sun_pos = (((double)GP142_YMAX - 30.0) - (HORIZON + 64)) / STAGES;
	isun_pos = (((double)GP142_YMAX - 30.0) - sun_y) / sun_pos;
	return (isun_pos);
}

/* the initial information the cars start with */
void initial_info(int y[MAX], int speed[MAX], int direction[MAX], 
				  int moving[MAX], int color[MAX], int i) {
	y[i] = HIGHWAY_CENTER - 10;
	direction[i] = +1;
	speed[i] = 5;
	color[i] = RED;
	moving[i] = TRUE;
}

void traffic(int direction[MAX], int x[MAX], int y[MAX],
			 int num_cars) {
	int i, n;
	for (i = 0; i <= num_cars; i++) {
		if (direction[i] == 1) {
			for (n = 0; n <= num_cars; n++) {
				if (direction[n] == 1 && n != i) {
					if (x[i] >= x[n] - 2 * CAR_WIDTH &&
						x[i] < x[n]) {
						x[i] = x[n] - 2 * CAR_WIDTH;
					}
				}
			}
		}
		else if(direction[i] == -1) {
			for (n = 0; n <= num_cars; n++) {
				if (direction[n] == -1 && n != i) {
					if (x[i] <= x[n] + 2 * CAR_WIDTH &&
						x[i] > x[n]) {
						x[i] = x[n] + 2 * CAR_WIDTH;
					}
				}
			}
		}
	}
}

/*****************************************************************************

     OTHER

*****************************************************************************/


/* Initialize random number generator */
/* [Ignore or delete if your program  */
/* doesn't user random numbers.] */
void initialize_random_numbers() {
	time_t t;			/* current time */
	srand (time(&t));
}



/* Display error message if internal error is detected. */
void something_is_wrong (void) {
	GP142_printfXY (RED, 0, 0, 16, "BAAAADDD NEWS!");
}